@@ -10,17 +10,78 @@ import os
 from urllib.parse import urlunparse, urlparse, quote
 import aiohttp
 from asyncio import Lock
+from copy import copy
 
 from tornado import gen, web, httpclient, httputil, process, websocket, ioloop, version_info
 
 from jupyter_server.utils import ensure_async, url_path_join
 from jupyter_server.base.handlers import JupyterHandler, utcnow
+from traitlets.traitlets import HasTraits
+from traitlets import Bytes, Dict, Instance, Integer, Unicode, Union, default, observe
 
 from .utils import call_with_asked_args
 from .websocket import WebSocketHandlerMixin, pingable_ws_connect
 from simpervisor import SupervisedProcess
 
 
+class RewritableResponse(HasTraits):
+    """
+    A class to hold the response to be rewritten by rewrite_response
+    """
+    # The following should not be modified (or even accessed) by rewrite_response.
+    # It is used to initialize the default values of the traits.
+    orig_response = Instance(klass=httpclient.HTTPResponse)
+
+    # The following are modifiable by rewrite_response
+    headers = Union(trait_types=[Dict(), Instance(klass=httputil.HTTPHeaders)])
+    body = Bytes()
+    code = Integer()
+    reason = Unicode(allow_none=True)
+
+    @default('headers')
+    def _default_headers(self):
+        return copy(self.orig_response.headers)
+
+    @default('body')
+    def _default_body(self):
+        return self.orig_response.body
+
+    @default('code')
+    def _default_code(self):
+        return self.orig_response.code
+
+    @default('reason')
+    def _default_reason(self):
+        return self.orig_response.reason
+
+    @observe('code')
+    def _observe_code(self, change):
+        # HTTP status codes are mapped to short descriptions in the
+        # httputil.responses dictionary, 200 maps to "OK", 403 maps to
+        # "Forbidden" etc.
+        #
+        # If code is updated and it previously had a reason matching its short
+        # description, we update reason to match the new code's short
+        # description.
+        #
+        if self.reason == httputil.responses.get(change['old'], 'Unknown'):
+            self.reason = httputil.responses.get(change['new'], 'Unknown')
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        # Trigger the default value to be set from orig_response on instantiation.
+        # Otherwise _observe_code will receive change['old'] == 0.
+        self.code
+
+    def _apply_to_copy(self, func):
+        """
+        Apply a function to a copy of self, and return the copy
+        """
+        new = copy(self)
+        func(new)
+        return new
+
+
 class AddSlashHandler(JupyterHandler):
     """Add trailing slash to URLs that need them."""
     @web.authenticated
@@ -44,9 +105,19 @@ class ProxyHandler(WebSocketHandlerMixin, JupyterHandler):
         self.proxy_base = ''
         self.absolute_url = kwargs.pop('absolute_url', False)
         self.host_allowlist = kwargs.pop('host_allowlist', ['localhost', '127.0.0.1'])
+        self.rewrite_response = kwargs.pop(
+            'rewrite_response',
+            tuple(),
+        )
         self.subprotocols = None
         super().__init__(*args, **kwargs)
 
+    # Support/use jupyter_server config arguments allow_origin and allow_origin_pat
+    # to enable cross origin requests propagated by e.g. inverting proxies.
+
+    def check_origin(self, origin=None):
+        return JupyterHandler.check_origin(self, origin)
+
     # Support all the methods that tornado does by default except for GET which
     # is passed to WebSocketHandlerMixin and then to WebSocketHandler.
 
@@ -180,6 +251,7 @@ class ProxyHandler(WebSocketHandlerMixin, JupyterHandler):
 
         req = httpclient.HTTPRequest(
             client_uri, method=self.request.method, body=body,
+            decompress_response=False,
             headers=headers, **self.proxy_request_options())
         return req
 
@@ -227,6 +299,7 @@ class ProxyHandler(WebSocketHandlerMixin, JupyterHandler):
         req = self._build_proxy_request(host, port, proxied_path, body)
 
         try:
+            # Here, "response" is a tornado.httpclient.HTTPResponse object.
             response = await client.fetch(req, raise_error=False)
         except httpclient.HTTPError as err:
             # We need to capture the timeout error even with raise_error=False,
@@ -249,19 +322,62 @@ class ProxyHandler(WebSocketHandlerMixin, JupyterHandler):
             self.set_status(500)
             self.write(str(response.error))
         else:
-            self.set_status(response.code, response.reason)
+            # Represent the original response as a RewritableResponse object.
+            original_response = RewritableResponse(orig_response=response)
+            
+            # The function (or list of functions) which should be applied to modify the
+            # response.
+            rewrite_response = self.rewrite_response
+
+            # If this is a single function, wrap it in a list.
+            if isinstance(rewrite_response, (list, tuple)):
+                rewrite_responses = rewrite_response
+            else:
+                rewrite_responses = [rewrite_response]
+
+            # To be passed on-demand as args to the rewrite_response functions.
+            optional_args_to_rewrite_function = {
+                'request': self.request,
+                'orig_response': original_response,
+                'host': host,
+                'port': port,
+                'path': proxied_path
+            }
+
+            # Initial value for rewriting
+            rewritten_response = original_response
+
+            for rewrite in rewrite_responses:
+                # The rewrite function is a function of the RewritableResponse object
+                # ``response`` as well as several other optional arguments. We need to
+                # convert it to a function of only ``response`` by plugging in the
+                # known values for all the other parameters. (This is called partial
+                # evaluation.)
+                def rewrite_pe(rewritable_response: RewritableResponse):
+                    return call_with_asked_args(
+                        rewrite,
+                        {
+                            'response': rewritable_response,
+                            **optional_args_to_rewrite_function
+                        }
+                    )
+                # Now we can cleanly apply the partially evaulated function to a copy of
+                # the rewritten response.
+                rewritten_response = rewritten_response._apply_to_copy(rewrite_pe)
+
+            ## status
+            self.set_status(rewritten_response.code, rewritten_response.reason)
 
             # clear tornado default header
             self._headers = httputil.HTTPHeaders()
-
-            for header, v in response.headers.get_all():
+            for header, v in rewritten_response.headers.get_all():
                 if header not in ('Content-Length', 'Transfer-Encoding',
-                                  'Content-Encoding', 'Connection'):
+                                  'Connection'):
                     # some header appear multiple times, eg 'Set-Cookie'
                     self.add_header(header, v)
 
-            if response.body:
-                self.write(response.body)
+            if rewritten_response.body:
+                self.write(rewritten_response.body)
 
     async def proxy_open(self, host, port, proxied_path=''):
         """
@@ -282,9 +398,7 @@ class ProxyHandler(WebSocketHandlerMixin, JupyterHandler):
             proxied_path = '/' + proxied_path
 
         client_uri = self.get_client_uri('ws', host, port, proxied_path)
-        headers = self.request.headers
-        current_loop = ioloop.IOLoop.current()
-        ws_connected = current_loop.asyncio_loop.create_future()
+        headers = self.proxy_request_headers()
 
         def message_cb(message):
             """
@@ -316,15 +430,13 @@ class ProxyHandler(WebSocketHandlerMixin, JupyterHandler):
             self.ws = await pingable_ws_connect(request=request,
                 on_message_callback=message_cb, on_ping_callback=ping_cb,
                 subprotocols=self.subprotocols)
-            ws_connected.set_result(True)
             self._record_activity()
             self.log.info('Websocket connection established to {}'.format(client_uri))
 
-        current_loop.add_callback(start_websocket_connection)
         # Wait for the WebSocket to be connected before resolving.
         # Otherwise, messages sent by the client before the
         # WebSocket successful connection would be dropped.
-        await ws_connected
+        await start_websocket_connection()
 
     def proxy_request_headers(self):
         '''A dictionary of headers to be used when constructing
@@ -365,6 +477,10 @@ class LocalProxyHandler(ProxyHandler):
     A tornado request handler that proxies HTTP and websockets
     from a port on the local system. Same as the above ProxyHandler,
     but specific to 'localhost'.
+
+    The arguments "port" and "proxied_path" in each method are extracted from
+    the URL as capture groups in the regex specified in the add_handlers
+    method.
     """
     async def http_get(self, port, proxied_path):
         return await self.proxy(port, proxied_path)
@@ -397,6 +513,10 @@ class RemoteProxyHandler(ProxyHandler):
     """
     A tornado request handler that proxies HTTP and websockets
     from a port on a specified remote system.
+
+    The arguments "host", "port" and "proxied_path" in each method are
+    extracted from the URL as capture groups in the regex specified in the
+    add_handlers method.
     """
 
     async def http_get(self, host, port, proxied_path):
@@ -478,7 +598,7 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
         url = 'http://localhost:{}'.format(self.port)
         async with aiohttp.ClientSession() as session:
             try:
-                async with session.get(url) as resp:
+                async with session.get(url, allow_redirects=False) as resp:
                     # We only care if we get back *any* response, not just 200
                     # If there's an error response, that can be shown directly to the user
                     self.log.debug('Got code {} back from {}'.format(resp.status, url))
@@ -565,17 +685,56 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
         return self.proxy(self.port, path)
 
 
-def setup_handlers(web_app, host_allowlist):
-    host_pattern = '.*$'
+def setup_handlers(web_app, serverproxy_config):
+    host_allowlist = serverproxy_config.host_allowlist
+    rewrite_response = serverproxy_config.non_service_rewrite_response
     web_app.add_handlers('.*', [
-        (url_path_join(web_app.settings['base_url'], r'/proxy/([^/]*):(\d+)(.*)'),
-         RemoteProxyHandler, {'absolute_url': False, 'host_allowlist': host_allowlist}),
-        (url_path_join(web_app.settings['base_url'], r'/proxy/absolute/([^/]*):(\d+)(.*)'),
-         RemoteProxyHandler, {'absolute_url': True, 'host_allowlist': host_allowlist}),
-        (url_path_join(web_app.settings['base_url'], r'/proxy/(\d+)(.*)'),
-         LocalProxyHandler, {'absolute_url': False}),
-        (url_path_join(web_app.settings['base_url'], r'/proxy/absolute/(\d+)(.*)'),
-         LocalProxyHandler, {'absolute_url': True}),
+        (
+            url_path_join(
+                web_app.settings['base_url'],
+                r'/proxy/([^/]*):(\d+)(.*)',
+            ),
+            RemoteProxyHandler,
+            {
+                'absolute_url': False,
+                'host_allowlist': host_allowlist,
+                'rewrite_response': rewrite_response,
+            }
+        ),
+        (
+            url_path_join(
+                web_app.settings['base_url'],
+                r'/proxy/absolute/([^/]*):(\d+)(.*)',
+            ),
+            RemoteProxyHandler,
+            {
+                'absolute_url': True,
+                'host_allowlist': host_allowlist,
+                'rewrite_response': rewrite_response,
+            }
+        ),
+        (
+            url_path_join(
+                web_app.settings['base_url'],
+                r'/proxy/(\d+)(.*)',
+            ),
+            LocalProxyHandler,
+            {
+                'absolute_url': False,
+                'rewrite_response': rewrite_response,
+            },
+        ),
+        (
+            url_path_join(
+                web_app.settings['base_url'],
+                r'/proxy/absolute/(\d+)(.*)',
+            ),
+            LocalProxyHandler,
+            {
+                'absolute_url': True,
+                'rewrite_response': rewrite_response,
+            },
+        ),
     ])
 
 # vim: set et ts=4 sw=4:
